/**
 * OW2 Util
 * Copyright (C) 2008-2009 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: EventDispatcher.java 5071 2009-07-27 16:25:56Z eyindanga $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.event.impl;

import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.Semaphore;
import java.util.concurrent.SynchronousQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;

import org.ow2.util.event.api.EventPriority;
import org.ow2.util.event.api.IEvent;
import org.ow2.util.event.api.IEventDispatcher;
import org.ow2.util.event.api.IEventListener;
import org.ow2.util.event.api.IEventToken;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;

/**
 * Define methods needed to register listeners and dispatch events.
 * @author missonng
 */
public class EventDispatcher implements IEventDispatcher {
    /**
     * A specialized queue synchronized give job to workers.
     * @author missonng
     */
    private final class EventQueue {
        /**
         * The maximum time to wait on the synchronous queue before retrying on the concurrentLinkedQueue.
         */
        private static final long SYNC_POLL_TIME = 1000;

        /**
         * A SynchronousQueue for worker to wait jobs.
         */
        private final SynchronousQueue<EventToken> synchronousQueue = new SynchronousQueue<EventToken>(true);

        /**
         * A ConcurrentLinkedQueue to store jobs when no worker is available.
         */
        private final ConcurrentLinkedQueue<EventToken> concurrentLinkedQueue = new ConcurrentLinkedQueue<EventToken>();

        /**
         * Try to give a job to a worker directly through the SynchronousQueue.<br>
         * Store it in the ConcurrentLinkedQueue if no worker is available.
         * @param token The token to process.
         */
        public void put(final EventToken token) {
            if (!synchronousQueue.offer(token)) {
                concurrentLinkedQueue.offer(token);
            }
        }

        /**
         * Try to take a job from the ConcurrentLinkedQueue.<br>
         * Wait on the synchronousQueue if no job is available.
         * @return The runnable to execute.
         */
        public EventToken take() throws InterruptedException {
            EventToken token;
            while (true) {
                token = concurrentLinkedQueue.poll();
                if (token != null) {
                    return token;
                }

                token = synchronousQueue.poll(SYNC_POLL_TIME, TimeUnit.MILLISECONDS);
                if (token != null) {
                    return token;
                }
            }
        }
    }

    /**
     * A thread speciliazed to process event tokens.<br>
     * Workers get tokens to process from the dispatcher queue.<br>
     * A worker dies when the special runnable KILL is taken from the queue.<br>
     * This method avoid complex synchronization to kill workers properly.
     * @author missonng
     */
    private final class EventWorker extends Thread {
        /**
         * The worker main method.
         */
        @Override
        public void run() {
            try {
                logger.debug("EventWorker {0} started.", this);
                /**
                 * Set the name.
                 */
                setName("EventWorker");
                final EventQueue queue = EventDispatcher.this.queue;
                final IEventToken kill = EventDispatcher.this.kill;
                while (true) {
                    final EventToken token = queue.take();

                    if (token == kill) {
                        return;
                    }

                    try {
                        token.lock.acquireUninterruptibly();
                        if (token.killed) {
                            token.lock.release();
                            continue;
                        }
                        token.workers.add(this);
                        token.lock.release();

                        try {
                            while (true) {
                                final IEventListener listener = token.listeners.poll();

                                if (listener == null) {
                                    break;
                                }

                                EventPriority priority = listener.getPriority();
                                Semaphore semaphore = token.semaphores.get(priority);
                                semaphore.acquireUninterruptibly();
                                semaphore.release();
                                try {
                                    listener.handle(token.event);
                                } catch (Throwable fault) {
                                    logger.debug("Listener {0} failed due to an unknown error. {1}", listener, fault);
                                    continue;
                                } finally {
                                    int count = token.counts.get(priority).decrementAndGet();
                                    while (count == 0 && priority != null) {
                                        int nextOrdinal = priority.ordinal() + 1;
                                        priority = nextOrdinal < EventPriority.values().length ? EventPriority.values()[nextOrdinal] : null;
                                        semaphore = token.semaphores.get(priority);
                                        count = token.counts.get(priority).get();
                                        semaphore.release();
                                    }
                                }
                            }
                        } finally {
                            token.lock.acquireUninterruptibly();
                            token.workers.remove(this);
                            token.lock.release();
                        }
                    } catch (Throwable fault) {
                        logger.debug("EventWorker {0} failed due to an unknown error. {1}", this, fault);
                        token.terminate();
                    }
                }
            } catch (Throwable fault) {
                logger.debug("EventWorker {0} died due to an unknown error. {1}", this, fault);
                createWorkers(1);
            } finally {
                logger.debug("EventWorker {0} finished.", this);
            }
        }
    }

    /**
     * Define methods need to manage events processing.
     * @author missonng
     */
    private final class EventToken implements IEventToken {
        /**
         * The event to manage.
         */
        private IEvent event;

        /**
         * The token state.
         */
        private boolean killed = false;

        /**
         * A semaphore to protect the token state and workers list.
         */
        private Semaphore lock = new Semaphore(1, true);

        /**
         * The token workers list.
         */
        private LinkedList<EventWorker> workers = new LinkedList<EventWorker>();

        /**
         * Listeners to process.
         */
        private ConcurrentLinkedQueue<IEventListener> listeners = new ConcurrentLinkedQueue<IEventListener>();

        /**
         * Map priorities to a semaphore that is unlock when the priority is allow to be process.
         */
        private Map<EventPriority, Semaphore> semaphores = new HashMap<EventPriority, Semaphore>();

        /**
         * Map priorities to its number of listener.
         */
        private Map<EventPriority, AtomicInteger> counts = new HashMap<EventPriority, AtomicInteger>();

        /**
         * The event token constructor.<br>
         * Used to instanciate the kill token.
         */
        private EventToken() {

        }

        /**
         * The event token constructor.<br>
         * @param event The event to manage.
         */
        private EventToken(final IEvent event) {
            this.event = event;

            int permit = 1;
            for (EventPriority priority : EventPriority.values()) {
                int count = 0;
                this.semaphores.put(priority, new Semaphore(permit, true));
                for (IEventListener listener : EventDispatcher.this.listeners.get(priority)) {
                    if (event.checkPermission(listener) && listener.accept(event)) {
                        this.listeners.add(listener);
                        permit = 0;
                        count++;
                    }
                }
                this.counts.put(priority, new AtomicInteger(count));
            }
            this.semaphores.put(null, new Semaphore(permit, true));
            this.counts.put(null, new AtomicInteger(0));
        }

        /**
         * Terminate the event processing.
         */
        @SuppressWarnings("deprecation")
        public void terminate() {
            logger.debug("EventToken {0} terminating.", this);

            this.lock.acquireUninterruptibly();

            this.killed = true;

            EventDispatcher.this.createWorkers(this.workers.size());
            for (EventWorker worker : this.workers) {
                worker.stop();
            }

            for (Semaphore semaphore : this.semaphores.values()) {
                semaphore.release();
            }

            this.listeners.clear();

            this.lock.release();

            logger.debug("EventToken {0} terminated.", this);
        }

        /**
         * Allocate a new thread to process this event.
         */
        public void allocate() {
            EventDispatcher.this.queue.put(this);
        }

        /**
         * Wait for the event to be process.
         * @param timeout How long to wait for the event to be process.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final long timeout) throws InterruptedException {
            return join(EventPriority.values()[EventPriority.values().length - 1], timeout, false);
        }

        /**
         * Wait for the event to be process.
         * @param timeout How long to wait for the event to be process.
         * @param kill If true, kill the event processing if it not terminated before timeout.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final long timeout, final boolean kill) throws InterruptedException {
            return join(EventPriority.values()[EventPriority.values().length - 1], timeout, kill);
        }

        /**
         * Wait for the event to be process.
         * @param priority Only wait for the specified priority to be process.
         * @param timeout How long to wait for the event to be process.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final EventPriority priority, final long timeout) throws InterruptedException {
            return join(priority, timeout, false);
        }

        /**
         * Wait for the event to be process.
         * @param priority Only wait for the specified priority to be process.
         * @param timeout How long to wait for the event to be process.
         * @param kill If true, kill the event processing if it not terminated before timeout.
         * @return True if the event has terminate its process before the timeout, false otherwise.
         * @throws InterruptedException If the thread has been interrupted.
         */
        public boolean join(final EventPriority priority, final long timeout, final boolean kill) throws InterruptedException {
            int nextOrdinal = priority.ordinal() + 1;
            final EventPriority nextPriority = nextOrdinal < EventPriority.values().length ? EventPriority.values()[nextOrdinal] : null;
            final Semaphore semaphore = this.semaphores.get(nextPriority);
            final boolean ok = semaphore.tryAcquire(timeout, TimeUnit.MILLISECONDS);
            if (ok) {
                semaphore.release();
            } else if (kill) {
                terminate();
            }
            return ok;
        }
    }

    /**
     * The dispatcher default size.
     */
    private static final int DEFAULT_SIZE = 1;

    private Log logger = LogFactory.getLog(EventDispatcher.class);

    /**
     * The listener map.<br>
     * Map each priority to a list of listeners.
     */
    private final HashMap<EventPriority, CopyOnWriteArrayList<IEventListener>> listeners;

    /**
     * The token queue.
     */
    private final EventQueue queue;

    /**
     * The special token used to kill workers.
     */
    private final EventToken kill;

    /**
     * The dispatcher state.
     */
    private AtomicBoolean available;

    /**
     * The number of workers.
     */
    private AtomicInteger size;

    /**
     * The dispatcher constructor.
     */
    public EventDispatcher() {
        this.kill = new EventToken();
        this.queue = new EventQueue();

        this.available = new AtomicBoolean(false);

        this.size = new AtomicInteger(DEFAULT_SIZE);

        this.listeners = new HashMap<EventPriority, CopyOnWriteArrayList<IEventListener>>();
        for (EventPriority priority : EventPriority.values()) {
            this.listeners.put(priority, new CopyOnWriteArrayList<IEventListener>());
        }
    }

    /**
     * Start the dispatcher.
     */
    public void start() {
        final boolean old = this.available.getAndSet(true);
        if (!old) {
            this.createWorkers(this.size.get());
        }
    }

    /**
     * Stop the dispatcher.
     */
    public void stop() {
        final boolean old = this.available.getAndSet(false);
        if (old) {
            this.destroyWorkers(this.size.get());
        }
    }

    /**
     * Get the dispatcher state.
     * @return The dispatcher state.
     */
    public boolean isAvailable() {
        return this.available.get();
    }

    /**
     * Get the number of workers.
     * @return The number of workers.
     */
    public int getNbWorkers() {
        return this.size.get();
    }

    /**
     * Set the number of workers.
     * @param nbWorkers The number of workers.
     */
    public void setNbWorkers(final int nbWorkers) {
        final int diff = this.size.getAndSet(nbWorkers) - nbWorkers;
        if (this.available.get()) {
            if (diff > 0) {
                destroyWorkers(diff);
            } else {
                createWorkers(-diff);
            }
        }
    }

    /**
     * Add a listener.
     * @param listener The listener to add.
     */
    public void addListener(final IEventListener listener) {
        this.listeners.get(listener.getPriority()).add(listener);
    }

    /**
     * Remove a listener.
     * @param listener The listener to remove.
     */
    public void removeListener(final IEventListener listener) {
        this.listeners.get(listener.getPriority()).remove(listener);
    }

    /**
     * Dispatch this event.
     * @param event The event to dispatch.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event) {
        return dispatch(event, EventPriority.NONE, Long.MAX_VALUE, 1);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param priority the priority to synchronize with, or null to get the token immediately.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final EventPriority priority) {
        return dispatch(event, priority, Long.MAX_VALUE, 1);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param timeout The maximum time dispatch waits for synchronous listeners to process.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final long timeout) {
        return dispatch(event, EventPriority.NONE, timeout, 1);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param nbWorker The number of worker to allocate.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final int nbWorker) {
        return dispatch(event, EventPriority.NONE, Long.MAX_VALUE, nbWorker);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param priority the priority to synchronize with, or null to get the token immediately.
     * @param timeout The maximum time dispatch waits for synchronous listeners to process.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final EventPriority priority, final long timeout) {
        return dispatch(event, priority, timeout, 1);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param priority the priority to synchronize with, or null to get the token immediately.
     * @param nbWorker The number of worker to allocate.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final EventPriority priority, final int nbWorker) {
        return dispatch(event, priority, Long.MAX_VALUE, nbWorker);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param timeout The maximum time dispatch waits for synchronous listeners to process.
     * @param nbWorker The number of worker to allocate.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final long timeout, final int nbWorker) {
        return dispatch(event, EventPriority.NONE, timeout, nbWorker);
    }

    /**
     * Dispatch an event.
     * @param event The event to dispatch.
     * @param priority the priority to synchronize with, or null to get the token immediately.
     * @param timeout The maximum time dispatch waits for synchronous listeners to process.
     * @param nbWorker The number of worker to allocate.
     * @return A token to manage the event processing, or null if the dispatcher is not available.
     */
    public IEventToken dispatch(final IEvent event, final EventPriority priority, final long timeout, final int nbWorker) {
        if (!this.available.get()) {
            return null;
        }

        long end = System.currentTimeMillis() + timeout;
        if (end < 0) {
            end = Long.MAX_VALUE;
        }

        EventToken token = new EventToken(event);

        for (int i = 0; i < nbWorker; i++) {
            token.allocate();
        }

        if (priority != null) {
            while (System.currentTimeMillis() < end) {
                try {
                    token.join(priority, end - System.currentTimeMillis());
                    break;
                } catch (InterruptedException e) {
                    continue;
                }
            }
        }

        return token;
    }

    /**
     * A helper method to create workers.
     * @param nb The number of workers to create.
     */
    private void createWorkers(final int nb) {
        for (int i = 0; i < nb; i++) {
            EventWorker worker = new EventWorker();
            worker.setDaemon(true);
            worker.start();
        }
    }

    /**
     * A helper method to destroy workers.
     * @param nb The number of workers to destroy.
     */
    private void destroyWorkers(final int nb) {
        final EventToken kill = this.kill;
        final EventQueue queue = this.queue;
        for (int i = 0; i < nb; i++) {
            queue.put(kill);
        }
    }
}
